extends TextureRect


const BLACK_INC = 10  # 100
const BLACK_START = 5   # 20
const GRID_LOC = Vector2(55, 350)
const SIZE = Vector2(8, 10)
const THORN_COLOR = 0

onready var scoreboard = get_node('Hud/Scoreboard')
onready var sound_match = get_node('SoundMatch')
onready var sound_bad_match = get_node('SoundBadMatch')

var fall_down_timer:Timer
var over_grid:Node2D
var remove_matches_timer:Timer
var setup_board_timer:Timer
var swap_back_timer:Timer
var tokens:Node2D
var under_grid:Node2D

var bounds := Rect2(Vector2.ZERO, SIZE)
var hilighted
var moving := true
var mute := false
var score := 0
var swap_back_tokens


func _ready():
	randomize()
	var sd = randi()
#	sd = 2734040769
	print('seed: %d' % [sd])
	seed(sd)

#	dlog(str(OS.window_size))

	var scale:float = 558 / OS.window_size.x
	var ideal := OS.window_size
	ideal.y *= scale
	SIZE.y = round((ideal.y - GRID_LOC.y) / 64.0)
	SIZE.y = max(SIZE.y, 10)
#	dlog(str(SIZE))

	under_grid = Node2D.new()
	under_grid.position = GRID_LOC
	add_child(under_grid)
	move_child(under_grid, 0)

	tokens = Node2D.new()
	tokens.position = GRID_LOC
	add_child(tokens)
	move_child(tokens, 1)

	over_grid = Node2D.new()
	over_grid.position = GRID_LOC
	add_child(over_grid)
	move_child(over_grid, 2)

	make_timers()
	get_node('Hud/ButtonMute').pressed = not mute
	update_score()

	setup_board_timer.start()


func _unhandled_input(event):
	if event.is_action_pressed('ui_quit'):
		get_tree().quit()
	elif event.is_action_pressed('ui_click'):
		handle_click(get_global_mouse_position())

#func _process(delta):
#	pass


func add_thorn(vec:Vector2) -> void:
	var thorn:Obstacle = Obstacle.new(Obstacle.TYPE_THORN, vec)
	over_grid.add_child(thorn)


func begin_selection(tok:Token) -> void:
	if tok.can_match:
		tok.glow()
		hilighted = tok


func dlog(s:String) -> void:
	get_node('Debug').text += s + '\n'


func end_selection(tok:Token) -> void:
	if not is_instance_valid(hilighted):
		hilighted = null
		return

	if not tok.can_match:
		hilighted.unglow()
		hilighted = null
		return

	var hi_p:Vector2 = hilighted.grid_pos
	var t_p:Vector2 = tok.grid_pos

	hilighted.unglow()

	if hi_p.distance_squared_to(t_p) == 1.0:
		player_move(tok)

	hilighted = null


func fall_down():
	var changed := false
	var chance_black := BLACK_START - score / BLACK_INC
	chance_black = max(chance_black, 4)
#	dlog('chance of black: %d' % [chance_black])

	for x in SIZE.x:
		var start
		var rack := []

		for y in range(SIZE.y-1, -1, -1):
			var vec := Vector2(x, y)
			var tok:Token = get_token(vec)

			if tok == null:
				if start == null:
					start = y
					changed = true
			elif start != null:
				rack.push_back(tok)

		if start != null:
			var y2 := -2
			for y1 in range(start, -1, -1):
				var vec1 = Vector2(x, y1)
				var tok = rack.pop_front()
				if tok == null:
					tok = place_token(vec1, y2, chance_black)
					y2 -= 1
				else:
					tok.move_to(vec1)

	assert(tokens.get_child_count() == SIZE.x * SIZE.y)

	if changed:
		remove_matches_timer.start()
	else:
		spring_traps()
		moving = false


# Return the color of the token, or -1.
func get_tcolor(v:Vector2) -> int:
	var l = get_token(v)
	if l:
		return l.color
	return -1


func get_obstacles(v=null, typ:int=-1) -> Array:
	var ret := []
	for obs in over_grid.get_children():
		if obs is Obstacle and (v == null or obs.grid_pos == v):
			if typ == -1 or typ == obs.trap_type:
				ret.push_back(obs)
	for obs in under_grid.get_children():
		if obs is Obstacle and (v == null or obs.grid_pos == v):
			if typ == -1 or typ == obs.trap_type:
				ret.push_back(obs)
	return ret


func get_token(v:Vector2) -> Token:
	for tok in tokens.get_children():
		if tok is Token and tok.grid_pos == v:
			return tok
	return null


func get_token_by_coord(v:Vector2):
	for tok in tokens.get_children():
		if tok is Token and tok.at_coord(v):
			return tok


func handle_click(gp:Vector2) -> void:
	if moving:
		return

	var tok = get_token_by_coord(gp - GRID_LOC)

	if tok:
		if hilighted and is_instance_valid(hilighted):
			end_selection(tok)
		else:
			begin_selection(tok)


func make_a_timer(time, fun) -> Timer:
	var t:Timer = Timer.new()
	t.wait_time = time
	t.one_shot = true
	t.connect('timeout', self, fun)
	add_child(t)
	return t


func make_timers() -> void:
	fall_down_timer = make_a_timer(0.25, 'fall_down')
	remove_matches_timer = make_a_timer(0.5, 'remove_matches')
	setup_board_timer = make_a_timer(0.5, 'setup_board')
	swap_back_timer = make_a_timer(0.5, 'swap_back')


func matches() -> Array:
	# Return a list of matching tokens.
	var mout := []

	# Get every match, including sub-matches.
	var offs := [
		Vector2.LEFT,
		Vector2.UP,
		]
	for y in range(SIZE.y-1, -1, -1):
		for x in range(SIZE.x-1, -1, -1):
			var vec := Vector2(x, y)
			var c1 := get_tcolor(vec)
			for off in offs:
				var ts := {}
				ts[get_token(vec)] = true

				var v:Vector2 = vec
				while true:
					v += off
					if v.x < 0 or v.y < 0:
						break

					if c1 == get_tcolor(v) and get_token(v).can_match:
						ts[get_token(v)] = true
					else:
						break

				if len(ts) > 2:
					for tok in ts:
						tok.glow()
					mout.append(ts)

	for mat in mout:
		var cpmat = mat.duplicate(true)
		for tok in cpmat:
			if tok.explode_vert:
				for y in SIZE.y:
					var loc := Vector2(tok.grid_pos.x, y)
					var atok:Token = get_token(loc)
					if not mat.has(atok):
						atok.gives_points = false
						mat[atok] = true
						atok.glow()
			if tok.explode_horiz:
				for x in SIZE.x:
					var loc := Vector2(x, tok.grid_pos.y)
					var atok:Token = get_token(loc)
					if not mat.has(atok):
						atok.gives_points = false
						mat[atok] = true
						atok.glow()

	# Remove duplicate matches.
	for i in len(mout):
		for j in range(i+1, len(mout)):
			for t in mout[i]:
				if mout[j].get(t):
					for t2 in mout[i].keys():
						mout[j][t2] = true
					mout[i].clear()
					break

	# Remove empty matches.
	var mout2 := []
	for a in mout:
		if len(a) != 0:
			mout2.append(a)

	return mout2


func place_token(vec, start_y=-2, chance_black=1) -> Token:
	var tok := Token.new(-chance_black)
	tok.grid_pos = Vector2(vec.x, start_y)
	tokens.add_child(tok)
	tok.move_to(vec)
	return tok


func player_move(tok:Token):
	if not is_instance_valid(hilighted):
		hilighted = null
		return

	moving = true
	swap_tokens(hilighted, tok)

	var m = matches()
	if len(m) > 0:
		remove_matches_timer.start()
	else:
		swap_back_tokens = [hilighted, tok]
		swap_back_timer.start()


func mark_exploders(mat:Dictionary) -> void:
	# Mark a random token in a match as an exploder.

	var vert := false
	var horiz := false
	var keys:Array = mat.keys()
	for i in len(keys)-1:
		if keys[i].gives_points:
			if keys[i].grid_pos.x != keys[i+1].grid_pos.x:
				horiz = true
			if keys[i].grid_pos.y != keys[i+1].grid_pos.y:
				vert = true

	var spec = randi() % len(keys)
	var tok = keys[spec]

	# Remove the exploder from the match, so it
	#  won't be destroyed.
	mat.erase(tok)
	tok.unglow()
	if vert and horiz:
		tok.make_explode('cross')
	if vert:
		tok.make_explode('vert')
	else:
		tok.make_explode('horiz')


func remove_matches() -> void:
	# Don't remove anything until the tokens
	#  stop moving.
	if tweens_running():
		remove_matches_timer.start()
		return

	var changed := false
	var lost_points := false

	for mat in matches():
		var obs_match = {}
		var score_len = 0
		var pointless = 0
		var negative := 1  # i.e. not negative score
		var main_color := -1
		var problem := false

		for tok in mat:
			if tok.gives_points:
#				assert(main_color == -1 or main_color == tok.color)
				if not (main_color == -1 or main_color == tok.color):
					problem = true
				main_color = tok.color
				obs_match[tok] = true
				score_len += 1
			else:
				pointless += 1

		if problem:
			dlog('problem with match:')
			for tok in mat:
				dlog('%d %s' % [tok.color, tok.gives_points])

		changed = create_obstacle(obs_match)

		if score_len > 3:
			mark_exploders(mat)

		score += round(pow(2.0, score_len-2)) * negative
		score += 2 * pointless * negative
		update_score()

		for tok in mat:
			tok.die()
			changed = true

	if changed and not mute:
		if lost_points:
			sound_bad_match.play()
		else:
			sound_match.play()

	fall_down_timer.start()


func setup_board() -> void:
	# Place obstacles.

	var offs := [  # looking at previous tokens left and down
		Vector2.LEFT,  # This just makes the big while
		Vector2.LEFT * 2,  #  statement easier to read.
		Vector2.DOWN,
		Vector2.DOWN * 2,
		]
	for y in range(SIZE.y-1, -1, -1):
		for x in SIZE.x:
			var vec := Vector2(x, y)
			var tok := place_token(vec)
			var color := tok.color
			while (get_tcolor(vec + offs[0]) == color \
					and get_tcolor(vec + offs[1]) == color) \
					or (get_tcolor(vec + offs[2]) == color \
					and get_tcolor(vec + offs[3]) == color):
				color = tok.random_color()
			yield(get_tree(), 'idle_frame')

	assert(len(matches()) == 0)
	moving = false


func create_obstacle(mat:Dictionary) -> bool:
	# Replace a random token in a match
	#  with an obstacle.

	if len(mat) < 3:
		return false

	var keys = mat.keys()
	var color = keys[0].color
	var obs:Token = keys[randi() % len(keys)]
	var vec:Vector2 = obs.grid_pos

	if true or color == THORN_COLOR:
		add_thorn(vec)

	return true


func spring_traps() -> void:
	for obs in get_obstacles():
		obs.spring(get_token(obs.grid_pos))


func swap_back():
	var t = swap_back_tokens
	if t == null:
		return
	if not is_instance_valid(t[0]):
		return
	if not is_instance_valid(t[1]):
		return

	swap_tokens(t[0], t[1])
	moving = false
	return


func swap_tokens(tok1, tok2) -> void:
	var p1:Vector2 = tok1.grid_pos
	var p2:Vector2 = tok2.grid_pos
	tok1.move_to(p2)
	tok2.move_to(p1)


func tweens_running():
	for tok in tokens.get_children():
		if tok.tween.is_active():
			return true


func update_score() -> void:
	score = max(score, 0)
	scoreboard.update_score(score)


func _on_ButtonMute_toggled(button_pressed):
	mute = not button_pressed
