extends GridObject
class_name Token


var button_arrow_horiz:Sprite
var button_arrow_vert:Sprite
var button_hilight:Sprite
var button_sprite:Sprite
var button_thorn:Sprite
var glyph:Sprite
var timer:Timer

var can_match := true
var color:int = -1
var explode_horiz := false
var explode_vert := false
var gives_points := true
var skull := false

var colors := [
	Color(0, 0.8, 0),
	Color.orange,
	Color.red,
	Color.blue,
	Color.purple,
	Color(0.75, 0.75, 0),
	]

var MAX_COLOR = len(colors)

var bright_colors := [
	Color(0.5, 0.9, 0.5),
	Color(0.9, 0.7, 0.5),
	Color.lightcoral,
	Color.cornflower,
	Color.orchid,
	Color(0.85, 0.85, 0.5),
	]

var button_glyphs := [
	'res://art/glyph_dots1.png',
	'res://art/glyph_dots2.png',
	'res://art/glyph_dots3.png',
	'res://art/glyph_dots4.png',
	'res://art/glyph_dots5.png',
	'res://art/glyph_dots6.png',
	'res://art/glyph_skull.png',
	]


#func _ready():
#	pass


func _init(in_color:int = -1):
	if in_color >= 0:
		color = in_color
	else:
		color = (randi() % MAX_COLOR)

	make_button()
	make_button_arrows()
	make_button_highlight()
	make_button_glyph()
	make_thorn()
	make_timer()

	change_color(color)

	add_to_group('tokens', true)


func change_color(c:int) -> void:
	color = c
	button_sprite.modulate = colors[color]
	change_glyph()


func change_glyph() -> void:
	glyph.texture = load(button_glyphs[color])


func die() -> void:
	tween.interpolate_method(self, 'dying_flash',
			0.0, 1.0, 0.1, Tween.TRANS_CUBIC, Tween.EASE_IN)
	tween.start()
	state = 'flashing'
	timer.wait_time = 0.1
	timer.start()


func dying_flash(val:float):
	button_sprite.scale *= 1.0 + val
	button_sprite.modulate.r += 0.1
	button_sprite.modulate.g += 0.1
	button_sprite.modulate.b += 0.1
	button_hilight.visible = false
	button_sprite.modulate.a -= 0.1


func glow() -> void:
	button_sprite.modulate = bright_colors[color]


func handle_timer():
	if state == 'flashing':
		get_tree().call_group('tokens', 'near_match', grid_pos)
		get_tree().call_group('obstacles', 'near_match', grid_pos)
		queue_free()


func make_button() -> void:
	var btn := Sprite.new()
	btn.texture = load('res://art/button1_color.png')
	btn.scale = Vector2.ONE * 0.25
	bounds = btn.get_rect()
	bounds.size *= btn.scale
	TOKEN_SIZE = bounds.size
	bounds.position -= TOKEN_SIZE / 2.0
	btn.modulate = colors[color]
	add_child(btn)
	button_sprite = btn


func make_button_arrows() -> void:
	var b := Sprite.new()
	b.texture = load('res://art/button1_arrow_horiz.png')
	b.scale = Vector2.ONE * 0.25
	b.visible = false
	button_arrow_horiz = b
	add_child(b)

	b = Sprite.new()
	b.texture = load('res://art/button1_arrow_vert.png')
	b.scale = Vector2.ONE * 0.25
	b.visible = false
	button_arrow_vert = b
	add_child(b)


func make_button_glyph() -> void:
	var gly := Sprite.new()
	gly.modulate = Color(1, 1, 1, 0.75)
	gly.scale = Vector2.ONE * 0.25
	if button_glyphs[color]:
		gly.texture = load(button_glyphs[color])
	else:
		gly.visible = false
	add_child(gly)
	glyph = gly


func make_button_highlight() -> void:
	var shd := Sprite.new()
	shd.texture = load('res://art/button1_shade.png')
	shd.scale = Vector2.ONE * 0.25
	add_child(shd)
	button_hilight = shd


func make_explode(typ:String) -> void:
	if typ == 'vert':
		explode_vert = true
	elif typ == 'horiz':
		explode_horiz = true
	elif typ == 'cross':
		explode_vert = true
		explode_horiz = true

	if explode_horiz:
		button_arrow_horiz.visible = true
	if explode_vert:
		button_arrow_vert.visible = true


func make_thorn() -> void:
	var spr := Sprite.new()
	spr.texture = load('res://art/thorn01.png')
	add_child(spr)
	spr.visible = false
	button_thorn = spr


func make_timer():
	var t:Timer = Timer.new()
	t.wait_time = 1.0
	t.one_shot = true
	t.connect('timeout', self, 'handle_timer')
	add_child(t)
	timer = t


func near_match(pos:Vector2) -> void:
	var dist := 0.0
	if button_thorn.visible:
		dist = grid_pos.distance_squared_to(pos)
		if dist == 1.0:
			unthorn()


func random_color() -> int:
	var c = randi() % MAX_COLOR
	change_color(c)
	return c


func thorn() -> void:
	button_thorn.visible = true
	can_match = false


func unglow() -> void:
	button_sprite.modulate = colors[color]


func unskull() -> void:
	skull = false
	glyph.visible = false


func unthorn() -> void:
	button_thorn.visible = false
	can_match = true
